/*!
 * Copyright (c) 2010, 2012 aizo ag, Switzerland
 *
 * \author Johannes Winkelmann, johannes.winkelmann@aizo.com
 * \author Sergey Bostandzhyan, jin@dev.digitalstrom.org
 */

#include <iomanip>

#include <cstdlib>
#include <cstdio>
#include <cassert>
#include <sstream>

#include <digitalSTROM/dsuid.h>
#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>
#include <dsf.h>

#include "firmwareimageinfo.h"
#include "errormessage.h"
#include "connection.h"

//! The size of the firmware data chunks expected by the dSM
const uint32_t DSF::CHUNK_SIZE = 64;


/*!
 * \brief initialize the firmware upgrade transfer
 * \param firmwareImageInfo the image info of the firmware image to be sent
 * \return true on success
 */
bool DSF::initializeTransfer(Connection* connection, dsuid_t& dsuid,
                             const FirmwareImageInfo& firmwareImageInfo)
{
  int result = ImageTransfer_write_request(connection->handle(),
                                           dsuid,
                                           firmwareImageInfo.manufacturer,
                                           firmwareImageInfo.device,
                                           firmwareImageInfo.contentType,
                                           firmwareImageInfo.contentSize,
                                           firmwareImageInfo.contentCrc);
  if (result != 0) {
    printf("Res: %d\n", result);
  }
  return result == 0;
}


/*!
 * \brief send a chunk of data to the dSM
 * \param data pointer to the data
 * \param size data size
 * \param packetNumber the packet sequence number
 * \return the return code as reported by the dSM
 */
int DSF::sendChunk(Connection* connection, const dsuid_t& dsuid,
                   uint8_t* data, uint32_t packetNumber,
                   uint32_t* expectedPacketNumber)
{
  uint32_t receivedPacketNumber;
  int ret = ImageTransfer_data(connection->handle(), dsuid, packetNumber, data,
                               expectedPacketNumber, &receivedPacketNumber);

  return ret;
}


/*!
 * \brief check whether image arrived correctly on dSM
 * \return true on success
 */
bool DSF::verifyImage(Connection* connection, dsuid_t dsuid, uint32_t DeviceId)
{
  uint16_t readCRC;
  uint16_t computedCRC;
  int result = dSFImageMaintenance_check_image(connection->handle(), dsuid, DeviceId,
                                            &readCRC, &computedCRC);
  if (result != 0) {
    return false;
  }

  return readCRC == computedCRC;
}

/*!
 * \brief ask dSM to start update, i.e. flash firmware image
 * \return true on success
 */
bool DSF::startUpdate(Connection* connection, dsuid_t dsuid, uint32_t DeviceId)
{
  int cmp;
  if (dsuid_is_broadcast(&dsuid)) {
    cmp = -2;
  } else {
    cmp = 0;
  }
  return (dSFImageMaintenance_start_update(connection->handle(), dsuid, DeviceId) == cmp);
}

static bool sort_busMembers(const busMemberDevice_t &mem1, const busMemberDevice_t &mem2)
{
  if (mem1.type > mem2.type) {
    return false;
  }
  if (mem1.type < mem2.type) {
    return true;
  }
  for (size_t i = 0; i < sizeof(dsuid_t); i++) {
    if (mem1.dsuid.id[i] > mem2.dsuid.id[i]) {
      return false;
    }
    if (mem1.dsuid.id[i] < mem2.dsuid.id[i]) {
      return true;
    }
  }
  return false;
}

int DSF::getBusMembers(Connection* connection, std::list<busMemberDevice_t>  &device_list, bool dSFDsuidOnly)
{
  dsuid_t tmp_device_list[63];
  busMemberDevice_t tmp_bus_member;
  int ret;

  ret = DsmApiGetBusMembers(connection->handle(), tmp_device_list, 63);
  if (ret < 0) {
    return ret;
  }

  for (int i = 0; i < ret; i++) {
    if (BusMember_get_type(connection->handle(), tmp_device_list[i], &tmp_bus_member.type) != ERROR_OK) {
      if (DsmApiIsdSM(tmp_device_list[i])) {
        tmp_bus_member.type = BusMember_dSM11;
      } else {
        tmp_bus_member.type = BusMember_Unknown;
      }
    }
    if (dSFDsuidOnly && !(tmp_bus_member.type == BusMember_dSF25)) {

      continue;
    }
    tmp_bus_member.dsuid = tmp_device_list[i];
    device_list.push_back(tmp_bus_member);
  }

  device_list.sort(sort_busMembers);
  return device_list.size();
}

static char const* BusMemberToString(uint8_t type)
{
    switch ((BusMemberDevice_t)type) {
    case BusMember_ds485d: return "dS485d";
    case BusMember_ds485p: return "dS485p";
    case BusMember_dSM11 : return "dSM11";
    case BusMember_dSM12 : return "dSM12";
    case BusMember_vDSM  : return "vdSM";
    case BusMember_vDC   : return "vdC";
    case BusMember_dSF25 : return "dSF25";
    case BusMember_dSS   : return "dSS";
    case BusMember_dSA   : return "dSA";
    case BusMember_dSG   : return "dSG";
    default:  return "Unknown";
    }
}

bool DSF::listDevices(Connection* connection, std::list<std::string>& devices, bool dSMDsuidOnly)
{
  std::list<busMemberDevice_t> device_list;

  int ret = DSF::getBusMembers(connection, device_list, dSMDsuidOnly);
  if (ret < 0) {
    return false;
  }

  std::list<busMemberDevice_t>::iterator it;
  for (it = device_list.begin(); it != device_list.end(); it++) {
    std::ostringstream str;

    for (int j = 0; j < DSUID_SIZE; ++j) {
      str << (j ? ":" : "")
          << std::setw(2)
          << std::setfill('0')
          << std::hex
          << (int)(*it).dsuid.id[j];
    }

    if (!dSMDsuidOnly) {
      uint32_t hwVersion = 0;
      uint32_t armSwVersion = 0;
      uint32_t dspSwVersion = 0;
      uint32_t boosterArmSwVersion = 0;
      uint16_t apiVersion = 0;
      std::string name = "";

      str << " "
          << std::setiosflags(std::ios::left)
          << std::setw(8)
          << std::setfill(' ')
          << BusMemberToString((*it).type);

      if ((*it).type == BusMember_dSF25)
      {
        if (getVersion(connection, (*it).dsuid,
                       &hwVersion, &armSwVersion, &dspSwVersion, &boosterArmSwVersion, &apiVersion, name)) {
          str << " (hw:"  << std::setw(8) << std::setfill('0') << hwVersion
              << ", arm:" << std::setw(8) << std::setfill('0') << armSwVersion
              << ", dsp:" << std::setw(8) << std::setfill('0') << dspSwVersion
              << ", booster arm:" << std::setw(8) << std::setfill('0') << boosterArmSwVersion
              << ", api:" << std::setw(8) << std::setfill('0') << apiVersion
              << ", '" << name << "')";
        } else {
          str << " getVersion() failed";
        }
      }
    }

    devices.push_back(str.str());
  }
  return true;
}

bool DSF::getDeviceCount(Connection* connection, int *count) {
    std::list<busMemberDevice_t> device_list;

  if (count == NULL) {
    return false;
  }

  *count = 0;

  int ret = DSF::getBusMembers(connection, device_list, true);
  if (ret < 0) {
    return false;
  }

  *count = ret;

  return true;
}

bool DSF::getVersion(Connection* connection,
                     dsuid_t& dsuid,
                     uint32_t* hwVersion,
                     uint32_t* armSwVersion,
                     uint32_t* dspSwVersion,
                     uint32_t* boosterArmSwVersion,
                     uint16_t* apiVersion,
                     std::string& name)
{
  uint8_t nameBuf[21];
  std::ostringstream ss;

  int ret = dSFInfo(connection->handle(), dsuid,
                    hwVersion, armSwVersion, dspSwVersion, boosterArmSwVersion, apiVersion,
                    NULL, nameBuf);

  if (ret != 0) {
    return false;
  }

  for (int i = 0; (i < 21) && nameBuf[i]; ++i) {
    ss << nameBuf[i];
  }
  name = ss.str();

  return true;
}



/*!
 * \brief dump a uint8_t buffer
 * \param data data buffer
 * \param length data buffer length
 */
void DSF::dumpBuffer(uint8_t* data, uint32_t length)
{
  unsigned int line = 0;
  for (uint32_t i = 0; i < length; i++){
    if (i % 8 == 0) {
      if (i > 0) {
        printf("\n");
      }
      printf("%02d: ", ++line);
    }
    printf("%02x ", data[i]);
  }
  printf("\n");
}


bool DSF::parseDSUIDFromString(dsuid_t* dsuid, const std::string& dsuidString)
{
  if (dsuidString.size() != (DSUID_SIZE - 1) * 3 + 2) {
    return false;
  }

  for (int i = 0; i < DSUID_SIZE; ++i) {
    std::string sub = dsuidString.substr(i*3, i*3+2);
    dsuid->id[i] = strtol(sub.c_str(), NULL, 16);
  }

  return true;
}
