/*!
 * Copyright (c) 2011, 2012 aizo ag, Switzerland
 *
 * \author Johannes Winkelmann, johannes.winkelmann@aizo.com
 * \author Sergey Bostandzhyan, jin@dev.digitalstrom.org
 */

#include <cstdlib>
#include <stdio.h>

#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>
#include <dsf.h>
#include <dsfgroup.h>

#include "firmwarefile.h"
#include "messagewriter.h"
#include "gettext_helper.h"


DsfGroup::DsfGroup(Connection* connection, bool messages) :
    m_connection(connection), m_messages(messages)
{
  pthread_mutexattr_t attr;

  pthread_mutexattr_init(&attr);
  pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);

  int ret = pthread_mutex_init(&m_dsf_list_mutex, &attr);
  if (ret != 0) {
    printf("Could not initialize dSF list mutex!\n");
  }

  ret = pthread_mutex_init(&m_dsf_full_list_mutex, NULL);
  if (ret != 0) {
    printf("Could not initialize dSF full list mutex!\n");
  }

  pthread_mutexattr_destroy(&attr);
}

DsfGroup::~DsfGroup()
{
  pthread_mutex_destroy(&m_dsf_list_mutex);
  pthread_mutex_destroy(&m_dsf_full_list_mutex);
}

int DsfGroup::sizeRemaining()
{
  lockDsfList();
  size_t s = m_dsfs.size();
  unlockDsfList();
  return s;
}

int DsfGroup::size()
{
  lockDsfFList();
  size_t s = m_dsfs_full.size();
  unlockDsfFList();
  return s;
}

void DsfGroup::resetRebootFlags()
{
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();
  for (; it != m_dsfs.end(); it++) {
    (*it)->rebooted = false;
    (*it)->expectReboot = false;
  }
  unlockDsfList();
}

void DsfGroup::add(const std::string& dsuid)
{
  boost::shared_ptr<Dsf> dsf(new Dsf());
  dsf->dsuidString = dsuid;
  DSF::parseDSUIDFromString(&(dsf->dsuid), dsuid);
  dsf->ok = false;
  dsf->rebooted = false;
  dsf->expectReboot = false;
  lockDsfList();
  m_dsfs.push_back(dsf);
  unlockDsfList();
  lockDsfFList();
  m_dsfs_full.push_back(dsf);
  unlockDsfFList();
}

void DsfGroup::removeUpToDateDsfs(FirmwareFile& firmwareFile,
                                  std::list<std::string>& uptoDateDsfs,
                                  bool removeFailed, bool verbose)
{
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();
  while (it != m_dsfs.end()) {
    uint32_t hwVersion = 0;
    uint32_t armSwVersion = 0;
    uint32_t dspSwVersion = 0;
    uint32_t boosterArmSwVersion = 0;
    uint16_t apiVersion = 0;
    std::string name;

    if (DSF::getVersion(m_connection, (*it)->dsuid,
                         &hwVersion, &armSwVersion, &dspSwVersion, &boosterArmSwVersion, &apiVersion, name)) {
      bool swVersionMatch = false;
      bool hwVersionMatch = false;

      switch (firmwareFile.imageInfo().device) {
      case DSP_DSF25:
        if (firmwareFile.imageInfo().version == dspSwVersion) {
          swVersionMatch = true;
        }
        hwVersionMatch = true;
        break;
      case ARM_DSF25:
        if (firmwareFile.imageInfo().version == armSwVersion) {
          swVersionMatch = true;
        }
        hwVersionMatch = true;
        break;
      case BOOSTER_ARM_DSF25:
        if (firmwareFile.imageInfo().version == boosterArmSwVersion) {
          swVersionMatch = true;
        }
        hwVersionMatch = true;
        break;
      }
      if (swVersionMatch || !hwVersionMatch) {
        if (hwVersionMatch) {
          uptoDateDsfs.push_back((*it)->dsuidString.c_str());
        }
        setExpectRebootFlag((*it)->dsuidString, false);
        if (verbose) {
          if (!hwVersionMatch) {
            printf("dSF %s is another type.\n",
                   (*it)->dsuidString.c_str());
          } else if (swVersionMatch) {
            printf("dSF %s is already up to date.\n",
                   (*it)->dsuidString.c_str());
          }
        }
        it = m_dsfs.erase(it);
        continue;
      }

    } else {
      printf("dsf '%s' getVersion() failed\n", (*it)->dsuidString.c_str());
      // let's ignore errors here; we can still try to upgrade this dSF
      if (removeFailed) {
        setExpectRebootFlag((*it)->dsuidString, false);
        printf("Removing dSF %s because getVersion failed\n", (*it)->dsuidString.c_str());
        it = m_dsfs.erase(it);
        continue;
      }
    }

    ++it;
  }
  unlockDsfList();
}

int DsfGroup::initializeTransfer(FirmwareFile& firmwareFile)
{
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();
  while (it != m_dsfs.end()) {
    if (!DSF::initializeTransfer(m_connection, (*it)->dsuid, firmwareFile.imageInfo())) {
      printf("\nRemoving dsf '%s' after initializeTransfer() failed\n", (*it)->dsuidString.c_str());
      it = m_dsfs.erase(it);
      continue;
    }

    ++it;
  }
  unlockDsfList();

  ImageTransfer_data_response_callback_t dataResponse = DsfGroup::dataCallback;
  DsmApiCallback_t callback;
  callback.function = (void*)dataResponse;
  callback.arg = this;
  DsmApiSetCallback(m_connection->handle(), DS485_CONTAINER_RESPONSE,
                    IMAGE_TRANSFER, IMAGE_TRANSFER_DATA,
                    &callback, NULL);

  return sizeRemaining();
}

int DsfGroup::verifyData(uint32_t deviceId)
{
  MessageWriter::getInstance()->msg(_("Verifying uploaded firmware image..."));
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();
  while (it != m_dsfs.end()) {
    if (!DSF::verifyImage(m_connection, (*it)->dsuid, deviceId)) {
      printf("\nRemoving dsf '%s' after verifyData() failed\n", (*it)->dsuidString.c_str());
      it = m_dsfs.erase(it);
      continue;
    }

    ++it;
  }
  unlockDsfList();
  return sizeRemaining();
}

void DsfGroup::setExpectRebootFlag(std::string dsuid, bool flag)
{
  lockDsfFList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs_full.begin();
  while (it != m_dsfs_full.end()) {
    if ((*it)->dsuidString == dsuid) {
      (*it)->expectReboot = flag;
      break;
    }
    it++;
  }
  unlockDsfFList();
}

int DsfGroup::startUpdate(uint32_t deviceId)
{
  DsmApiCallback_t callback_struct;
  callback_struct.function = (void*)DsfGroup::busChangeCallback;
  callback_struct.arg = this;
  DsmApiSetBusChangeCallback(m_connection->handle(), &callback_struct, NULL);

  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();
  while (it != m_dsfs.end()) {
    if (!DSF::startUpdate(m_connection, (*it)->dsuid, deviceId)) {
      printf("Removing dsf '%s' after startUpdate() failed\n",
              (*it)->dsuidString.c_str());
      it = m_dsfs.erase(it);
      continue;
    } else {
      (*it)->rebooted = false;
      (*it)->expectReboot = true;
    }
    ++it;
  }
  unlockDsfList();
  m_returned = 0;
  return sizeRemaining();
}

void DsfGroup::reduce()
{
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();
  while (it != m_dsfs.end()) {
    if (!(*it)->ok) {
      setExpectRebootFlag((*it)->dsuidString, false);
      printf("\nRemoving dSF %s due to timeout\n", (*it)->dsuidString.c_str());
      it = m_dsfs.erase(it);
      continue;
    }

    ++it;
  }
  unlockDsfList();
}


void DsfGroup::containedDsuids(std::list<std::string>& dsmList)
{
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();
  for (; it != m_dsfs.end(); ++it) {
    dsmList.push_back((*it)->dsuidString);
  }
  unlockDsfList();
}

int DsfGroup::transferData(FirmwareFile& firmwareFile)
{
  uint32_t transferred = 0;
  int readBytes = 0;
  uint8_t buffer[DSF::CHUNK_SIZE];

  uint32_t expectedPacketNumber;
  uint32_t packetNumber;
  int32_t errorCode;

  packetNumber = 0;
  m_nextPacketNo = 0;
  
  std::string fwtype = FirmwareFile::deviceIdToString((DeviceId_t)firmwareFile.imageInfo().device);

  firmwareFile.setPosition(0);

  int last_progress = -1;

  while (transferred < firmwareFile.imageInfo().contentSize) {
    int progress = 100 -
      ((firmwareFile.imageInfo().contentSize - transferred) * 100 /
       firmwareFile.imageInfo().contentSize);
    if (((progress % 5) == 0) && (progress != last_progress)) {
      MessageWriter::getInstance()->msg(_P("dSF firmware upgrade in progress. "
                  "Uploading %s firmware, %d%% completed...", fwtype.c_str(),
                  progress));
      last_progress = progress;
    }
    readBytes = firmwareFile.readChunk(buffer, DSF::CHUNK_SIZE);
    if (readBytes == -1) {
      printf("\nError reading firmware file\n");
      unregisterDataCallback();
      return RETCODES::ERR_FILE_IO;
    }

    prepareForNextPacket();
    packetNumber = m_nextPacketNo;
    ++m_nextPacketNo; // might be modified in callback
    errorCode = DSF::sendChunk(m_connection, DSUID_BROADCAST,
                               buffer, packetNumber, &expectedPacketNumber);
    if (errorCode != -2) {
      printf("\nError in sendChunk(): %d\n", errorCode);
      unregisterDataCallback();
      return RETCODES::ERR_DSM_IMG_TRANSFER;
    }



    // we grant a 5 second timeout for the dSMs to call back; that's 100000 us * 50
    int sleepCount = 0;
    while (!allOk()) {
      // let the callbacks happen
      usleep(100000);
      ++sleepCount;
      if (sleepCount == 50) {
        // remove dSM's that haven't answered
        reduce();
        break;
      }
    }

    if (sizeRemaining() == 0) {
      printf("\nNo dSFs left; aborting\n");
      return RETCODES::ERR_GENERAL;
    }

    if (m_nextPacketNo > packetNumber) {
      transferred += readBytes;
      if (m_nextPacketNo % 20 == 0) {
        printf(".");
        fflush(stdout);
      }
    } else {
      // some dSF requested a retransmission
      firmwareFile.setPosition(m_nextPacketNo*DSF::CHUNK_SIZE);
    }
  }
  printf("\n");
  printf(">> Transferred %d bytes in %d packet%s\n",
         transferred, packetNumber, (packetNumber > 1 ? "s" : ""));

  unregisterDataCallback();
  // allow UI to pick up the 100% message
  if (m_messages) {
      usleep(5000000);
  }
  return RETCODES::SUCCESS;
}


void DsfGroup::dataCallback(uint8_t errorCode, void *arg,
                            dsuid_t sourceId, dsuid_t destinationId,
                            uint32_t expectedPacketNo, uint32_t receivedPacketNo)
{
  DsfGroup* dsmGroup = static_cast<DsfGroup*>(arg);
  dsmGroup->handleDataCallback(errorCode, sourceId, destinationId,
                               expectedPacketNo, receivedPacketNo);
}


void DsfGroup::handleDataCallback(uint8_t errorCode,
                                  dsuid_t sourceId, dsuid_t destinationId,
                                  uint32_t expectedPacketNo, uint32_t receivedPacketNo)
{
  (void)destinationId;
  (void)receivedPacketNo;
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();

  while (it != m_dsfs.end()) {
    if (dsuid_equal(&(*it)->dsuid, &sourceId)) {

      if (errorCode != ERROR_OK) {
        switch (errorCode) {
          case ERROR_WRONG_PACKET_NR:
            if (expectedPacketNo < m_nextPacketNo) {
              printf("Go back in fw: cur: %d new: %d\n", m_nextPacketNo, expectedPacketNo);
              m_nextPacketNo = expectedPacketNo;
            }
            break;
          default:
            setExpectRebootFlag((*it)->dsuidString, false);
            printf("\nRemoving dSF %s due to faulty callback (%d / 0x%x)\n",
                   (*it)->dsuidString.c_str(), errorCode, errorCode);
            it = m_dsfs.erase(it);
            continue;
        }
      }
      (*it)->ok = true;
    }

    ++it;
  }
  unlockDsfList();
}

void DsfGroup::prepareForNextPacket()
{
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();
  for (; it != m_dsfs.end(); ++it) {
    (*it)->ok = false;
  }
  unlockDsfList();
}


bool DsfGroup::allOk()
{
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::const_iterator it = m_dsfs.begin();
  for (; it != m_dsfs.end(); ++it) {
    if (!(*it)->ok) {
      unlockDsfList();
      return false;
    }
  }
  unlockDsfList();
  return true;
}

void DsfGroup::restore()
{
  lockDsfList();
  lockDsfFList();
  m_dsfs.assign(m_dsfs_full.begin(), m_dsfs_full.end());
  unlockDsfFList();
  unlockDsfList();
};

bool DsfGroup::online()
{
  std::list<boost::shared_ptr<Dsf> >::iterator it;
  uint32_t hwVersion = 0;
  uint32_t armSwVersion = 0;
  uint32_t dspSwVersion = 0;
  uint32_t boosterArmSwVersion = 0;
  uint16_t apiVersion = 0;
  std::string name;

  lockDsfList();
  for (it = m_dsfs.begin(); it != m_dsfs.end(); it++) {
    if (DSF::getVersion(m_connection, (*it)->dsuid,
                        &hwVersion, &armSwVersion, &dspSwVersion, &boosterArmSwVersion, &apiVersion,
                        name) == false) {
      unlockDsfList();
      return false;
    }
    usleep(500000); // wait half a second
  }
  unlockDsfList();

  return true;
};

bool DsfGroup::allRebooted() {
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();
  for (; it != m_dsfs.end(); it++) {
    if (((*it)->expectReboot == true) && ((*it)->rebooted == false)) {
      unlockDsfList();
      return false;
    }
  }
  unlockDsfList();

  return true;
}

void DsfGroup::unregisterBusChangeCallback()
{
  DsmApiCallback_t callback_struct;
  callback_struct.function = NULL;
  callback_struct.arg = NULL;
  DsmApiSetBusChangeCallback(m_connection->handle(), &callback_struct, NULL);
}

bool DsfGroup::waitForDevices()
{
#define MICROSECONDS         1000000
#define WAIT_FOR_BUS_RETRIES 5
#define SLEEP_BUS_ACTIVE     5
#define SLEEP_REBOOT         5
#define SLEEP_GET_MEMBERS    15
#define SLEEP_GET_INFO       15

  int x = sizeRemaining() + 10;
  if (sizeRemaining() > 40) {
      x = x * 2;
  }

  MessageWriter::getInstance()->msg(_("Please wait..."));
  long long timeoutSecs = ((x*x)/2);
  printf("Will wait %lld seconds for dSFs to rejoin.\n", timeoutSecs);
  printf("Waiting for bus to become active...");
  fflush(stdout);
  int retries = 0;
  while (!m_connection->getLinkState() && (retries < WAIT_FOR_BUS_RETRIES)) {
      // won't consider the timeout, go by count here
    usleep(SLEEP_BUS_ACTIVE * MICROSECONDS);
    retries++;
    printf(".");
    fflush(stdout);
  }
  printf(" done.\n");

  if (sizeRemaining() == 0) {
    unregisterBusChangeCallback();
    return true;
  }

  printf("Waiting for dSFs to get back online...");
  fflush(stdout);

  while (!allRebooted() && (timeoutSecs > 0)) {
    usleep(SLEEP_REBOOT * MICROSECONDS);
    printf("<");
    fflush(stdout);
    timeoutSecs = timeoutSecs - SLEEP_REBOOT;
    MessageWriter::getInstance()->msg(_P("Installing dSF firmware: %zd / "
                                         "%zd completed", m_returned,
                                         sizeRemaining()));
  }
  unregisterBusChangeCallback();
  resetRebootFlags();

  usleep(1 * MICROSECONDS);

  do {
    int count;
    if (DSF::getDeviceCount(m_connection, &count)) {
      if (count >= sizeRemaining()) {
        break;
      }
    }
    printf("#");
    fflush(stdout);
    MessageWriter::getInstance()->msg(_P("Installing dSF firmware: %zd / "
                                         "%zd completed", count,
                                         sizeRemaining()));
    usleep(SLEEP_GET_MEMBERS * MICROSECONDS);
    timeoutSecs = timeoutSecs - SLEEP_GET_MEMBERS;
  } while (timeoutSecs > 0);

  usleep(1 * MICROSECONDS);

  do {
    if (online()) {
      printf(" done.\n");
      return true;
    }
    printf("+");
    fflush(stdout);
    usleep(SLEEP_GET_MEMBERS * MICROSECONDS);
    timeoutSecs = timeoutSecs - SLEEP_GET_INFO;
  } while (timeoutSecs > 0);

  printf(" timeout!\n");
  // give UI time to pick up the last state
  if (m_messages) {
    usleep(5000000);
  }
  return false;
}

void DsfGroup::busChangeCallback(void *arg, dsuid_t *dsuid, int flag)
{
  DsfGroup* dsmGroup = static_cast<DsfGroup*>(arg);
  dsmGroup->handleBusChange(dsuid, flag);
}

void DsfGroup::handleBusChange(dsuid_t *dsuid, int flag) {
  // only interested in those who appear on the bus
  if (flag) {
    return;
  }
  lockDsfList();
  std::list<boost::shared_ptr<Dsf> >::iterator it = m_dsfs.begin();

  for (; it != m_dsfs.end(); it++) {
    if (dsuid_equal(&(*it)->dsuid, dsuid)) {
      (*it)->rebooted = true;
      m_returned++;
      break;
    }
  }
  unlockDsfList();
}

void DsfGroup::lockDsfList()
{
  int ret = pthread_mutex_lock(&m_dsf_list_mutex);
  if (ret != 0) {
    printf("Could not lock dSF list mutex!\n");
  }
}

void DsfGroup::unlockDsfList()
{
  int ret = pthread_mutex_unlock(&m_dsf_list_mutex);
  if (ret != 0) {
    printf("Could not unlock dSF list mutex!\n");
  }
}

void DsfGroup::lockDsfFList()
{
  int ret = pthread_mutex_lock(&m_dsf_full_list_mutex);
  if (ret != 0) {
    printf("Could not lock dSF full list mutex!\n");
  }
}

void DsfGroup::unlockDsfFList()
{
  int ret = pthread_mutex_unlock(&m_dsf_full_list_mutex);
  if (ret != 0) {
    printf("Could not unlock dSF full list mutex!\n");
  }
}

void DsfGroup::unregisterDataCallback()
{
  DsmApiCallback_t callback;
  callback.function = NULL;
  callback.arg = NULL;
  DsmApiSetCallback(m_connection->handle(), DS485_CONTAINER_RESPONSE,
                    IMAGE_TRANSFER, IMAGE_TRANSFER_DATA,
                    &callback, NULL);
}

