/*!
 * Copyright (c) 2010 aizo ag, Switzerland
 *
 * \author Johannes Winkelmann, johannes.winkelmann@aizo.com
 */

#include <sys/stat.h>
#include <fcntl.h>
#include <cstring>
#include <cerrno>
#include <stdint.h>
#include <unistd.h>

#include "firmwarefile.h"
#include "errormessage.h"


FirmwareFile::FirmwareFile(ErrorMessage& errorMessage)
  : m_errorMessage(errorMessage)
{
  init();
}

void FirmwareFile::init()
{
  m_opened = false;
  m_fd = -1;
  m_fileSize = 0;
  m_errorMessage.setError("File not open", RETCODES::ERR_GENERAL);
}


/*!
 * \brief open a firmware file, verify format, determine size
 * \param fileName file name of the firmware file
 * \return true on success
 *
 * this method opens the firmware file and parses the firmware
 * header. from the header, information like manufacturer ID and the
 * like as well as the CRC are extracted.
 */
bool FirmwareFile::open(const std::string& fileName)
{
  m_fd = ::open(fileName.c_str(), O_RDONLY);
  if (m_fd == -1) {
    m_errorMessage.setError(strerror(errno), RETCODES::ERR_FILE_OPEN);
    return false;
  }

  m_fileSize = ::lseek(m_fd, 0, SEEK_END);
  ::lseek(m_fd, 0, SEEK_SET);

  if (!parseHeader()) {
    return false;
  }

  /* seek to actual data */
  ::lseek(m_fd, m_imageInfo.headerSize, SEEK_SET);

  m_opened = true;
  return true;
}

/*!
 * \brief read a chunk of data from the file
 * \param buffer to save data to
 * \param count size of data buffer
 * \return -1 on failure, number of bytes read otherwise
 */
int FirmwareFile::readChunk(uint8_t* buffer, unsigned int count)
{
  if (!m_opened) {
    return -1;
  }

  ssize_t len = read(m_fd, buffer, count);
  if (len == -1) {
    m_errorMessage.setError(strerror(errno), RETCODES::ERR_FILE_IO);
  }
  return len;
}

/*!
 * \brief set the file position within the firmware image; ignore header
 * \param pos position within the file
 * \return true on success
 */
bool FirmwareFile::setPosition(unsigned int pos)
{
  return ::lseek(m_fd, m_imageInfo.headerSize+pos, SEEK_SET) != -1;
}


/*!
 * \brief close the firmware file
 */
void FirmwareFile::close()
{
  ::close(m_fd);
  init();
}


/*!
 * \return the size of the file
 * \deprecated will be replaced by some image info later
 */
const FirmwareImageInfo& FirmwareFile::imageInfo() const
{
  return m_imageInfo;
}


/*!
 * parse firmware image header
 */
bool FirmwareFile::parseHeader()
{
  uint8_t buffer[4];

  if (read(m_fd, buffer, 4) != 4) {
    return false;
  }
  if (read(m_fd, &m_imageInfo.headerSize, 4) != 4) {
    return false;
  }
  if (read(m_fd, &m_imageInfo.manufacturer, 4) != 4) {
    return false;
  }
  if (read(m_fd, &m_imageInfo.device, 4) != 4) {
    return false;
  }
  if (read(m_fd, &m_imageInfo.contentType, 4) != 4) {
    return false;
  }
  if (read(m_fd, &m_imageInfo.contentSize, 4) != 4) {
    return false;
  }
  if (read(m_fd, &m_imageInfo.contentCrc, 2) != 2) {
    return false;
  }

  // read version, which is after the magic tag
  uint32_t dummy;
  if (read(m_fd, &dummy, 2) != 2) { // reserved
    return false;
  }
  if (read(m_fd, &dummy, 4) != 4) { // magic tag: 0xdeadbeef
    return false;
  }
  if (read(m_fd, &m_imageInfo.version, 4) != 4) {
    return false;
  }

  // - verify format
  if (memcmp(buffer, FIRMWARE_HEADER_TAG, 4) != 0) {
    m_errorMessage.setError("Invalid firmware image format", RETCODES::ERR_GENERAL);
    return false;
  }

  if ((m_imageInfo.headerSize + m_imageInfo.contentSize) != m_fileSize) {
    m_errorMessage.setError("Invalid content size", RETCODES::ERR_GENERAL);
    return false;
  }

  return true;
}

std::string FirmwareFile::deviceIdToString(DeviceId_t deviceId)
{
  switch (deviceId) {
  case ARM_DSF25:
    return std::string("dSF25 ARM");
  case BOOSTER_ARM_DSF25:
    return std::string("dSF25 Booster ARM");
  case DSP_DSF25:
    return std::string("dSF25 DSP");
  default:
    return std::string("");
  }
}
