/*!
 * Copyright (c) 2010, 2012 aizo ag, Switzerland
 *
 * \author Johannes Winkelmann, johannes.winkelmann@aizo.com
 * \author Sergey Bostandzhyan, jin@dev.digitalstrom.org
 */

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <iostream>
#include <iomanip>
#include <cstring>

#include <cstdio>
#include <cstdlib>
#include <cerrno>
#include <iterator>

#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>

#include "firmwareupgrade.h"
#include "firmwarefile.h"
#include <dsf.h>
#include "errormessage.h"
#include <dsfgroup.h>
#include "connection.h"
#include "messagewriter.h"
#include "gettext_helper.h"

using std::cout;
using std::endl;
using std::string;

const std::string FirmwareUpgrade::DEFAULT_CONNECTION = "tcp://localhost:8442";

FirmwareUpgrade::FirmwareUpgrade() : m_debug(false), m_pretend(false)
{
}

/*!
 * \brief print application usage
 */
void FirmwareUpgrade::usage()
{
  cout << "Usage:" << endl;

  cout << "> List:             "
       << "  " << basename(m_appName) << " -l -c <connection>" << endl;
  cout << "> Update:           "
       << "  " << basename(m_appName)
       << "    -c <connection> -d <dsf dsuid> <image file(s)>" << endl;
  cout << "> Broadcast Update: "
       << "  " << basename(m_appName)    << "    -c <connection> -B <image file(s)>" << endl;
  cout << "\n"
       << "  Options:\n"
       << "    -D  print debug output\n"
       << "    -p  pretend; only list dSFs that would be updated\n"
       << "    -m  <msgfile> write update status messages to given file\n"
       << "    -f  upgrade even when update and device versions match"
       << endl;
}

bool FirmwareUpgrade::parseArguments(int argc, char** argv)
{
  int opt;
  while ((opt = getopt(argc, argv, "Bc:Dd:m:lhpVvf")) != -1) {
    switch (opt) {
      case 'c':
        m_connectionString = optarg;
        break;
      case 'd':
        m_dsfDSUID = optarg;
        break;
      case 'D':
        m_debug = true;
        break;
      case 'l':
        m_scanMode = true;
        break;
      case 'B':
        m_broadcastMode = true;
        break;
      case 'p':
        m_pretend = true;
        break;
      case 'V':
        m_versionMode = true;
        break;
      case 'm':
        m_msgFile = optarg;
        break;
      case 'f':
        m_force = true;
        break;
      case 'h':
        usage();
        exit(0);
      case 'v':
        /* we print the version anyway, so just exit */
        exit(0);
      default: /* '?' */
        return false;
    }
  }

  // we expect maximum three non-option arguments, DSP and ARM and ARM-Booster image files
  while (optind < argc) {
    m_fileNames.push_back(std::string(argv[optind]));
    optind++;
  }

  if ((m_scanMode && m_versionMode) ||
      (m_scanMode && m_broadcastMode) ||
      (m_versionMode && m_broadcastMode)) {
    return false;
  }

  if (m_scanMode) {
    // nothing more needed
  } else if (m_versionMode) {
    if (m_dsfDSUID.empty()) {
      return false;
    }
  } else {

    if (m_fileNames.empty()) {
      return false;
    }

    if (!m_broadcastMode && m_dsfDSUID.empty()) {
      return false;
    }
  }

  return true;
}


/*!
 * \brief entry point for the upgrade
 * \param argc application argument count
 * \param argv application argument vector
 * \return RETCODES::SUCCESS on success, error code otherwise
 *
 * \see RETCODES::ReturnValue
 */
int FirmwareUpgrade::run(int argc, char** argv)
{
  m_appName = argv[0];
  int ret = RETCODES::SUCCESS;

  cout << ">> dSF Firmware upgrade v" << VERSION << endl;

  // - defaults
  m_scanMode = false;
  m_versionMode = false;
  m_broadcastMode = false;
  m_force = false;
  if (!parseArguments(argc, argv)) {
    usage();
    return RETCODES::ERR_ARG;
  }

  ErrorMessage m_connectionError;
  if (m_connectionString.empty()) {
    m_connectionString = DEFAULT_CONNECTION;
    if (m_debug) {
      cout << "No connnection specified, defaulting to " << m_connectionString << endl;
    }
  }

  m_connection = new Connection(m_connectionError);
  if (!m_connection->open(m_connectionString)) {
    cout << "Failed to initialize ds485 lib on " << m_connectionString << endl;
    cout << "Error: " << m_connectionError.message << endl;
    return RETCODES::ERR_DSM_LIB_INIT;
  }

  MessageWriter *writer = MessageWriter::createInstance(m_msgFile);
  if (writer == NULL) {
    cout << "Coult not initialize message writer!" << endl;
    m_connection->close();
    return RETCODES::ERR_GENERAL;
  }

#define MAX_RETRIES 20
  int retries = 0;
  cout << "Waiting for bus to become ACTIVE..." << endl;

  MessageWriter::getInstance()->msg(_("Waiting for dS485 bus to become "
                                      "active... "));
  while (!m_connection->getLinkState() && (retries < MAX_RETRIES)) {
    usleep(5000000); // microseconds, 5s
    retries++;
  }

  if (retries < MAX_RETRIES) {
    cout << "Bus ACTIVE, proceeding." << endl;
  } else {
    // retries >= 10, means we did not manage to become active, and there is
    // no point to go any further without an active bus
    cout << "Bus not ACTIVE, exiting." << endl;
    MessageWriter::getInstance()->msg(_("dS485 bus not active, please try "
                                        "again later!"));
    m_connection->close();
    return RETCODES::ERR_DSM_CONNECT;
  }

  // wait one token round time 10ms * 64 devices, round this up to one second
  usleep(1000000); // microseconds, 1s

  if (m_scanMode) {
    ret = listDevices();
  } else if (m_versionMode) {
    ret = getDSFVersion();
  } else if (m_broadcastMode) {
    ret = broadcastUpdate();
  } else {
    ret = updateFirmware();
  }
  m_connection->close();

  delete writer;

  return ret;
}

int FirmwareUpgrade::listDevices()
{
  std::list<string> devices;
  DSF::listDevices(m_connection, devices, false);

  cout << "Available devices:" << endl;
  std::list<std::string>::iterator it;
  int i = 1;
  for (it = devices.begin(); it != devices.end(); ++it, ++i) {
    cout << std::setw(2) << i << ")  " << *it << endl;
  }
  return RETCODES::SUCCESS;
}

int FirmwareUpgrade::getDSFVersion()
{
  dsuid_t dsuid;

  if (!DSF::parseDSUIDFromString(&dsuid, m_dsfDSUID)) {
    cout << "Error: invalid dsuid: " << m_dsfDSUID << endl;
    return -1;
  }

  uint32_t hwVersion = 0;
  uint32_t armSwVersion = 0;
  uint32_t dspSwVersion = 0;
  uint32_t boosterArmSwVersion = 0;
  uint16_t apiVersion = 0;
  string name;
  if (!DSF::getVersion(m_connection, dsuid,
                       &hwVersion, &armSwVersion, &dspSwVersion, &boosterArmSwVersion, &apiVersion, name)) {
    cout << "Error while reading dsf version" << endl;
    return -1;
  }

  cout << std::hex
       << hwVersion << " " << armSwVersion << dspSwVersion << " " << apiVersion
       << " " << name << endl;
  return RETCODES::SUCCESS;
}

int FirmwareUpgrade::broadcastUpdate()
{
  std::list<string> devices;
  DSF::listDevices(m_connection, devices, true);

  DsfGroup group(m_connection, !m_msgFile.empty());
  std::list<std::string>::iterator it;
  for (it = devices.begin(); it != devices.end(); ++it) {
    group.add(*it);
  }

  return upgradeGroup(&group);
}

int FirmwareUpgrade::updateFirmware()
{
  DsfGroup group(m_connection, !m_msgFile.empty());
  group.add(m_dsfDSUID);

  return upgradeGroup(&group);
}

bool FirmwareUpgrade::executeUpgrade(DsfGroup* group, FirmwareFile *fwFile)
{
  group->initializeTransfer(*fwFile);
  if (group->transferData(*fwFile) != RETCODES::SUCCESS) {
    cout << "Failed to transfer data to the dSFs" << endl;
    MessageWriter::getInstance()->msg(_("Failed to transfer firmware!"));
    return false;
  }
  group->verifyData(fwFile->imageInfo().device);
  group->startUpdate(fwFile->imageInfo().device);

  return true;
}

void FirmwareUpgrade::updateSummary(DsfGroup* group, time_t *start, time_t *end,
                                    std::list<std::string>& updatedDsms)
{
  std::list<std::string> devices;
  group->containedDsuids(devices);
  if (group->sizeRemaining() > 0) {
    cout << "Successfully updated " << group->sizeRemaining() << " "
         << (group->sizeRemaining() == 1 ? "dSF" : "dSFs") << ": " << endl;

    std::list<std::string>::iterator it;
    for (it = devices.begin(); it != devices.end(); ++it) {
      cout << "  " << *it << endl;
      updatedDsms.push_back(*it);
    }
  }

  uint32_t minutes = (*end-*start) / 60;
  uint32_t seconds = (*end-*start) - minutes*60;
  cout << "Update took "
       << minutes << " min "
       << std::setfill('0') << std::setw(2) << seconds
       << " " << (seconds == 1 ? "second" : "seconds")
       << endl;
}

// 0 == error
// 0x 0000 0001 = dSM1x / DSP
// 0x 0000 0002 = dSM11 / ARM
// 0x 0000 0003 = dSM12 / ARM
// 0x 0000 0100 = dSM20 / ARM
// 0x 0000 0101 = dSM20 / DSP
// 0x 0000 0200 = dSF25 / ARM
// 0x 0000 0201 = dSF25 / DSP
// 0x 0000 0202 = dSF25 / booster-ARM
uint32_t FirmwareUpgrade::getFirmwareType(std::string fileName)
{
  ErrorMessage errorMessage;
  uint32_t ret = 0;
  if (!fileName.empty()) {
    FirmwareFile tempfw(errorMessage);
    if (!tempfw.open(fileName)) {
      cout << "Error in '" << fileName << "': " << errorMessage.message << endl;
    } else {
      ret = tempfw.imageInfo().device;
      tempfw.close();
    }
  }
  return ret;
}

void FirmwareUpgrade::sortFirmwareFiles(std::list<std::string>& fileNameList)
{
  std::list<std::string> tempList;
  std::list<std::string>::iterator it;
  for (it = fileNameList.begin(); it != fileNameList.end(); ++it) {
    uint32_t type = getFirmwareType(*it);
    if (type == DSP_DSF25 || type == BOOSTER_ARM_DSF25) {
      tempList.push_back(*it);
    } else if (ARM_DSF25) {
      tempList.push_front(*it);
    } else {
      cout << "Warning: Firmware in'" << *it
          << "' incompatible with dSF hardware so ignoring." << endl;
    }
  }
  fileNameList.swap(tempList);
}

int FirmwareUpgrade::upgradeGroup(DsfGroup* group)
{
  uint32_t fwType;
  std::list<std::string> upToDateDsfs;
  std::list<std::string> updatedARM;
  std::list<std::string> updatedDSP;
  std::list<std::string> updatedBoosterARM;
  int retcode = RETCODES::SUCCESS;

  sortFirmwareFiles(m_fileNames);

  std::list<std::string>::iterator fwIter;
  for (fwIter = m_fileNames.begin(); fwIter != m_fileNames.end(); ++fwIter) {
    cout << "Downloading image " << *fwIter << "..." << endl;

    fwType = getFirmwareType(*fwIter);
    if (fwType == ARM_DSF25) {
      // ARM images
      retcode = upgradeGroupFirmware(group, *fwIter, upToDateDsfs, updatedARM);
    } else if (fwType == DSP_DSF25) {
      // DSP image
      retcode = upgradeGroupFirmware(group, *fwIter, upToDateDsfs, updatedDSP);
    } else if (fwType == BOOSTER_ARM_DSF25) {
      // DSP image
      retcode = upgradeGroupFirmware(group, *fwIter, upToDateDsfs, updatedBoosterARM);
    }

    if (retcode != RETCODES::SUCCESS) {
      return retcode;
    }

    upToDateDsfs.clear();
  }

  upToDateDsfs.assign(updatedARM.begin(), updatedARM.end());
  upToDateDsfs.insert(upToDateDsfs.end(), updatedDSP.begin(), updatedDSP.end());
  upToDateDsfs.insert(upToDateDsfs.end(), updatedBoosterARM.begin(), updatedBoosterARM.end());
  upToDateDsfs.sort();
  upToDateDsfs.erase(std::unique(upToDateDsfs.begin(), upToDateDsfs.end()),
                     upToDateDsfs.end());
  return upToDateDsfs.size();
}

int FirmwareUpgrade::upgradeGroupFirmware(DsfGroup* group, std::string fileName,
                                          std::list<std::string>& upToDateDsfs,
                                          std::list<std::string>& updatedDsfs)
{
  ErrorMessage errorMessage;
  FirmwareFile firmwareFile(errorMessage);
  time_t start, end;
  bool upgradeOK = false;
  int retcode = RETCODES::SUCCESS;
  string fwtype;

  if (!firmwareFile.open(fileName)) {
    cout << "Error in '" << fileName << "': " << errorMessage.message << endl;
    return errorMessage.value;
  }

  FirmwareImageInfo imageInfo = firmwareFile.imageInfo();

  fwtype = FirmwareFile::deviceIdToString((DeviceId_t)imageInfo.device);

  if (m_debug) {
    printf("HS: %d, Man: %x, Dev: %d, CT: %x, CS: %x, CC: %x, V: %08x\n",
           imageInfo.headerSize, imageInfo.manufacturer, imageInfo.device,
           imageInfo.contentType, imageInfo.contentSize, imageInfo.contentCrc,
           imageInfo.version);
  }

  if (!m_force) {
    group->removeUpToDateDsfs(firmwareFile, upToDateDsfs);
  }

  if (m_debug) {
    std::list<std::string>::iterator it = upToDateDsfs.begin();
    for (; it != upToDateDsfs.end(); ++it) {
      cout << "dsf '" << *it << "' already up to date" << endl;
    }
  }

  if (group->sizeRemaining() > 0) {
    cout << "Upgrading " << group->sizeRemaining() << " "
         << (group->sizeRemaining() == 1 ? "dSF" : "dSFs")
         << endl;
  } else {
    cout << "No dSFs to upgrade." << endl << endl;
  }

  if (group->sizeRemaining() == 0) {
    firmwareFile.close();
    cout << "Removing image files from dSFs" << endl;
    group->restore();
    return RETCODES::SUCCESS;
  }

  int retriesRemaining = MAX_TRIES;

  while (group->sizeRemaining() != 0 && retriesRemaining > 0) {
    time(&start);
    if (!m_pretend) {
      upgradeOK = executeUpgrade(group, &firmwareFile);
    } else {
      upgradeOK = true;
    }
    time(&end);

    if (upgradeOK) {
      updateSummary(group, &start, &end, updatedDsfs);
    }
  
    group->restore();
    if (group->sizeRemaining() > 0) {
      cout << endl << "Waiting for " << group->sizeRemaining() << " "
           << (group->sizeRemaining() == 1 ? "dSF" : "dSFs")
           << " to rejoin..." << endl << endl;
    }

    if (!group->waitForDevices()) {
      cout << "Warning: not all of the updated dSFs are back online!" << endl;
    }

    upToDateDsfs.clear();
    group->removeUpToDateDsfs(firmwareFile, upToDateDsfs, true, false);

    retriesRemaining--;

    if (m_pretend) {
      break;
    }

    if (group->sizeRemaining() != 0 && retriesRemaining > 0) {
        std::list<std::string> devices;
        std::list<std::string>::iterator it;
        MessageWriter::getInstance()->msg(_P("Retrying %s update...", fwtype.c_str()));
        cout << "Retrying upgrade on following "
             << (group->sizeRemaining() == 1 ? "dSF" : "dSFs") << ": "
             << endl;
        group->containedDsuids(devices);
        for (it = devices.begin(); it != devices.end(); ++it) {
          cout << "  " << *it << endl;
        }
        devices.clear();
    }

  }

  if (!m_pretend) {
    if (group->sizeRemaining() != 0) {
      std::list<std::string> devices;
      std::list<std::string>::iterator it;
      cout << "The following "
           << (group->sizeRemaining() == 1 ? "dSF" : "dSFs")
           << " could not be upgraded: " << endl;
      group->containedDsuids(devices);
      for (it = devices.begin(); it != devices.end(); ++it) {
        cout << "  " << *it << endl;
      }
      retcode = RETCODES::ERR_MISSED_UPDATE;
    }
  }

  firmwareFile.close();
  cout << "Removing image files from dSFs" << endl;
  group->restore();

  return retcode;
}

